#  Copyright 2020 Marco Köpcke (Parakoopa)
#
#  This file is part of py-desmume.
#
#  py-desmume is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  py-desmume is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with py-desmume.  If not, see <https://www.gnu.org/licenses/>.
import os
from typing import Dict, Optional, List

from gi.repository import Gtk, Gdk

from desmume.controls import key_names, Keys, key_names_localized
from desmume.frontend.widget_to_primitive import widget_to_primitive


class KeyboardControlsDialogController:
    """This dialog shows the keyboard controls."""
    def __init__(self, parent_window: Gtk.Window):
        path = os.path.abspath(os.path.dirname(__file__))
        # SkyTemple translation support
        try:
            from skytemple.core.ui_utils import make_builder
            self.builder = make_builder(os.path.join(path, "PyDeSmuMe_controls.glade"))
        except ImportError:
            self.builder = Gtk.Builder()
            self.builder.add_from_file(os.path.join(path, "PyDeSmuMe_controls.glade"))
        self.window: Gtk.Window = self.builder.get_object('wKeybConfDlg')
        self.window.set_transient_for(parent_window)
        self.window.set_attached_to(parent_window)
        self._keyboard_cfg: Optional[List[int]] = None
        self._tmp_key = None
        self.builder.connect_signals(self)

    def run(self, keyboard_cfg: List[int]) -> Optional[List[int]]:
        """Configure the keyboard configuration provided using the dialog,
        returns the new keyboard config if changed, else None."""
        self._keyboard_cfg = keyboard_cfg.copy()
        for i in range(0, Keys.NB_KEYS):
            b = self.builder.get_object(f"button_{key_names[i]}")
            b.set_label(f"{key_names_localized[i]} : {Gdk.keyval_name(self._keyboard_cfg[i])}")
        response = self.window.run()

        self.window.hide()
        if response == Gtk.ResponseType.OK:
            return self._keyboard_cfg

    # KEYBOARD CONFIG / KEY DEFINITION
    def on_wKeyDlg_key_press_event(self, widget: Gtk.Widget, event: Gdk.EventKey, *args):
        self._tmp_key = event.keyval
        self.builder.get_object("label_key").set_text(Gdk.keyval_name(self._tmp_key))
        return True

    def on_button_kb_key_clicked(self, w, *args):
        key = widget_to_primitive(w)
        dlg = self.builder.get_object("wKeyDlg")
        key -= 1  # key = bit position, start with
        self._tmp_key = self._keyboard_cfg[key]
        self.builder.get_object("label_key").set_text(Gdk.keyval_name(self._tmp_key))
        if dlg.run() == Gtk.ResponseType.OK:
            self._keyboard_cfg[key] = self._tmp_key
            self.builder.get_object(f"button_{key_names[key]}").set_label(f"{key_names_localized[key]} : {Gdk.keyval_name(self._tmp_key)}")

        dlg.hide()

    # Joystick configuration / Key definition
    def on_button_joy_key_clicked(self, w, *args):
        pass  # not part of this

    def gtk_widget_hide_on_delete(self, w: Gtk.Widget, *args):
        w.hide_on_delete()
        return True



"""
Some simple ports from sdlcntrl.cpp that are generally useful for dealing with controls.

Some of this only applies if used with GTK (eg. the keyboard configuration), other things
apply generally, such as the joystick configuration.
"""
#  Copyright 2020 Marco Köpcke (Parakoopa)
# 
#  This file is part of py-desmume.
# 
#  py-desmume is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
# 
#  py-desmume is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with py-desmume.  If not, see <https://www.gnu.org/licenses/>.
from typing import Tuple, List, TYPE_CHECKING
from desmume.i18n_util import _


if TYPE_CHECKING:
    from desmume.emulator import DeSmuME


class Joy:
    """Joystick input types."""
    JOY_AXIS = 0
    JOY_HAT = 1
    JOY_BUTTON = 2


class JoyHats:
    """Jostick hat identifiers."""
    JOY_HAT_RIGHT = 0
    JOY_HAT_LEFT = 1
    JOY_HAT_UP = 2
    JOY_HAT_DOWN = 3


class Keys:
    """DS key identifiers. NB_KEYS contains the total number of keys."""
    NB_KEYS	= 15
    KEY_NONE = 0
    KEY_A = 1
    KEY_B = 2
    KEY_SELECT = 3
    KEY_START = 4
    KEY_RIGHT = 5
    KEY_LEFT = 6
    KEY_UP = 7
    KEY_DOWN = 8
    KEY_R = 9
    KEY_L = 10
    KEY_X = 11
    KEY_Y = 12
    KEY_DEBUG = 13
    KEY_BOOST = 14
    KEY_LID = 15
    NO_KEY_SET = 0xFFFF


# Indices are emulator keys (see above), starting with KEY_A.
# Values are Gdk key codes and SDL joystick codes respectively.
# Keys.NO_KEY_SET is a magic number for no key set (sint16 -1)
default_config_keyboard = [
    120, 122, 65506, 65293, 65363, 65361, 65362, 65364, 119, 113, 115, 97, 112, 111, 65288
]
default_config_joystick = [
    513, 512, 517, 520, 1, 0, 2, 3, 519, 518, 516, 515, Keys.NO_KEY_SET, Keys.NO_KEY_SET, 514
]

key_names = [
    "A", "B", "Select", "Start",
    "Right", "Left", "Up", "Down",
    "R", "L", "X", "Y",
    "Debug", "Boost",
    "Lid"
]


key_names_localized = [
    _("A"),  # TRANSLATORS: DS Key name
    _("B"),  # TRANSLATORS: DS Key name
    _("Select"),  # TRANSLATORS: DS Key name
    _("Start"),  # TRANSLATORS: DS Key name
    _("Right"),  # TRANSLATORS: DS Key name
    _("Left"),  # TRANSLATORS: DS Key name
    _("Up"),  # TRANSLATORS: DS Key name
    _("Down"),  # TRANSLATORS: DS Key name
    _("R"),  # TRANSLATORS: DS Key name
    _("L"),  # TRANSLATORS: DS Key name
    _("X"),  # TRANSLATORS: DS Key name
    _("Y"),  # TRANSLATORS: DS Key name
    _("Debug"),  # TRANSLATORS: DS Key name
    _("Boost"),  # TRANSLATORS: DS Key name
    _("Lid")  # TRANSLATORS: DS Key name
]


def add_key(keypad, key):
    """
    Add a key to a keypad -> press the key. ``key`` is the keymask returned by ``keymask``.

    You don't need to call this manually, see
    :func:`~desmume.module.emulator.DeSmuME_Input.keypad_add_key` instead.
    """
    return keypad | key


def rm_key(keypad, key):
    """
    Remove a key from a keypad -> release the key. ``key`` is the keymask returned by ``keymask``.

    You don't need to call this manually, see
    :func:`~desmume.module.emulator.DeSmuME_Input.keypad_rm_key` instead.
    """
    return keypad & ~key


def keymask(k):
    """Returns the keymask for key ``k``. ``k`` is a constant of the ``Keys`` class."""
    return 1 << (k - 1) if k > 0 else 0


def load_default_config() -> Tuple[List[int], List[int]]:
    """
    Returns the default (keyboard configuration),(joystick configuration).

    The keyboard configuration is Gdk key IDs.
    """
    return default_config_keyboard, default_config_joystick


def load_configured_config(emu: 'DeSmuME') -> Tuple[List[int], List[int]]:
    """
    Load the default for inputs.
    Also set's the default config for joystick in emulator.

    The keyboard configuration is Gdk key IDs.

    :todo: Support loading/saving from the DesMuME config file.
    """
    kbcfg, jscfg = load_default_config()

    for i, jskey in enumerate(jscfg):
        emu.input.joy_set_key(i, jskey)

    return kbcfg, jscfg



"""
Software based rendering implementation (using Cairo). Can also be used directly,
it has a hooking mechanism for drawing custom overlays.
"""
#  Copyright 2020 Marco Köpcke (Parakoopa)
#
#  This file is part of py-desmume.
#
#  py-desmume is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  py-desmume is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with py-desmume.  If not, see <https://www.gnu.org/licenses/>.
from math import radians
from typing import Callable

import cairo
from gi.repository import Gtk, Gdk

from desmume.emulator import SCREEN_HEIGHT, SCREEN_WIDTH, DeSmuME, SCREEN_PIXEL_SIZE
from desmume.frontend.gtk_drawing_area_desmume import AbstractRenderer


class SoftwareRenderer(AbstractRenderer):

    def __init__(self, emu: DeSmuME, after_render_hook: Callable[[cairo.Context, int], None] = None):
        """The after rendering hook takes the ctx and display_id as params."""
        super().__init__(emu)
        self._upper_image = None
        self._lower_image = None
        self._after_render_hook = after_render_hook
        self.decode_screen()

    def init(self):
        pass

    def screen(self, base_w, base_h, ctx: cairo.Context, display_id: int):
        if display_id == 0:
            self.decode_screen()

        ctx.translate(base_w * self._scale / 2, base_h * self._scale / 2)
        ctx.rotate(-radians(self._screen_rotation_degrees))
        if self._screen_rotation_degrees == 90 or self._screen_rotation_degrees == 270:
            ctx.translate(-base_h * self._scale / 2, -base_w * self._scale / 2)
        else:
            ctx.translate(-base_w * self._scale / 2, -base_h * self._scale / 2)
        ctx.scale(self._scale, self._scale)
        if display_id == 0:
            ctx.set_source_surface(self._upper_image)
        else:
            ctx.set_source_surface(self._lower_image)
        ctx.get_source().set_filter(cairo.Filter.NEAREST)
        ctx.paint()

        if self._after_render_hook:
            self._after_render_hook(ctx, display_id)

    def reshape(self, draw: Gtk.DrawingArea, display_id: int):
        pass

    def decode_screen(self):
        gpu_framebuffer = self._emu.display_buffer_as_rgbx()

        self._upper_image = cairo.ImageSurface.create_for_data(
            gpu_framebuffer[:SCREEN_PIXEL_SIZE*4], cairo.FORMAT_RGB24, SCREEN_WIDTH, SCREEN_HEIGHT
        )

        self._lower_image = cairo.ImageSurface.create_for_data(
            gpu_framebuffer[SCREEN_PIXEL_SIZE*4:], cairo.FORMAT_RGB24, SCREEN_WIDTH, SCREEN_HEIGHT
        )



#  Copyright 2020 Marco Köpcke (Parakoopa)
#
#  This file is part of py-desmume.
#
#  py-desmume is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  py-desmume is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with py-desmume.  If not, see <https://www.gnu.org/licenses/>.

import os

import gi

from desmume.emulator import DeSmuME
from desmume.frontend.controller import MainController

gi.require_version('Gtk', '3.0')

from gi.repository import Gtk, GLib
from gi.repository.Gtk import Window


def main():
    path = os.path.abspath(os.path.dirname(__file__))

    # Load desmume
    emu = DeSmuME()

    # Init joysticks
    emu.input.joy_init()

    # Load Builder and Window
    builder = Gtk.Builder()
    builder.add_from_file(os.path.join(path, "PyDeSmuMe.glade"))
    main_window: Window = builder.get_object("wMainW")
    main_window.set_role("PyDeSmuME")
    GLib.set_application_name("PyDeSmuME")
    GLib.set_prgname("pydesmume")
    # TODO: Deprecated but the only way to set the app title on GNOME...?
    main_window.set_wmclass("PyDeSmuME", "PyDeSmuME")

    # Load main window + controller
    MainController(builder, main_window, emu)

    main_window.present()
    Gtk.main()
    del emu

    # Uninit SDL
    # TODO


if __name__ == '__main__':
    # TODO: At the moment the demo frontend doesn't support any cli arguments.
    main()



"""
Code for rendering the emulator on a Gtk DrawingArea. Feel free to use this in your own implementations!
Uses the OpenGL screen texture if possible, otherwise renders via Software.
"""
#  Copyright 2020 Marco Köpcke (Parakoopa)
#
#  This file is part of py-desmume.
#
#  py-desmume is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  py-desmume is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with py-desmume.  If not, see <https://www.gnu.org/licenses/>.
from abc import ABC, abstractmethod

import gi

gi.require_version('Gtk', '3.0')

import cairo
from gi.repository import Gtk

from desmume.emulator import DeSmuME


class AbstractRenderer(ABC):
    def __init__(self, emu: DeSmuME):
        self._emu = emu
        self._screen_rotation_degrees = 0
        self._scale = 1.0

    @classmethod
    def impl(cls, emu: DeSmuME) -> 'AbstractRenderer':
        from desmume.frontend.gtk_drawing_impl.opengl import OpenGLRenderer
        from desmume.frontend.gtk_drawing_impl.software import SoftwareRenderer
        # TODO: Implement OpenGL renderer
        if False and emu.has_opengl():
            return OpenGLRenderer(emu)
        else:
            return SoftwareRenderer(emu)

    @abstractmethod
    def init(self):
        pass

    @abstractmethod
    def screen(self, base_w, base_h, ctx: cairo.Context, display_id: int):
        pass

    @abstractmethod
    def reshape(self, draw: Gtk.DrawingArea, display_id: int):
        pass

    def set_scale(self, value):
        self._scale = value

    def get_scale(self):
        return self._scale

    def get_screen_rotation(self):
        return self._screen_rotation_degrees

    def set_screen_rotation(self, value):
        self._screen_rotation_degrees = value



"""
It is 2020 and Glade still doesn't support passing anything but Glade-defined objects as user data.
The GTK2 C++ Glade frontend had a custom Glade "parser" that worked around this, but we don't.
So instead, enjoy this!
"""
#  Copyright 2020 Marco Köpcke (Parakoopa)
#
#  This file is part of py-desmume.
#
#  py-desmume is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  py-desmume is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with py-desmume.  If not, see <https://www.gnu.org/licenses/>.
from gi.repository import Gtk


def widget_to_primitive(w: Gtk.Widget):
    name = Gtk.Buildable.get_name(w)
    if name.startswith("%d:"):
        return int(name[3:])
    elif name.startswith("%f:"):
        return float(name[3:])
    raise ValueError("Invalid widget for widget_to_primitive.")



#  Copyright 2020 Marco Köpcke (Parakoopa)
#
#  This file is part of py-desmume.
#
#  py-desmume is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  py-desmume is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with py-desmume.  If not, see <https://www.gnu.org/licenses/>.
import cairo
import gi

from desmume.controls import Keys, keymask, load_default_config, key_names, load_configured_config
from desmume.emulator import DeSmuME, NB_STATES, SCREEN_WIDTH, SCREEN_HEIGHT, StartFrom, DeSmuME_Date
from desmume.frontend.control_ui.joystick_controls import JoystickControlsDialogController
from desmume.frontend.control_ui.keyboard_controls import KeyboardControlsDialogController
from desmume.frontend.gtk_drawing_area_desmume import AbstractRenderer
from desmume.frontend.widget_to_primitive import widget_to_primitive

gi.require_version('Gtk', '3.0')

from gi.repository import Gtk, Gdk, GLib
from gi.repository.Gtk import *


TICKS_PER_FRAME = 17


class MainController:
    def __init__(self, builder: Builder, window: Window, emu: DeSmuME):
        self.builder = builder
        self.window = window
        self.emu = emu

        self.renderer = AbstractRenderer.impl(emu)
        self.renderer.init()

        self._keyboard_cfg, self._joystick_cfg = load_configured_config(emu)
        self._boost = False
        self._save_fs = 0
        self._frameskip = 0
        self._boost_fs = 20
        self._registered_main_loop = False
        self._fs_frame_count = 0
        self._fps_frame_count = 0
        self._fps_sec_start = 0
        self._fps = 0
        self._ticks_prev_frame = 0
        self._ticks_cur_frame = 0
        self._fps_limiter_disabled = False
        self._screen_invert = False
        self._click = False
        self._screen_right_force = False
        self._current_screen_width = SCREEN_WIDTH
        self._current_screen_height = SCREEN_HEIGHT
        self._screen_gap = False
        self._screen_no_gap = False
        self._screen_right = False
        self._supress_event = False

        self._filter_nds = Gtk.FileFilter()
        self._filter_nds.set_name("Nintendo DS ROMs (*.nds)")
        self._filter_nds.add_pattern("*.nds")

        self._filter_gba_ds = Gtk.FileFilter()
        self._filter_gba_ds.set_name("Nintendo DS ROMs with binary loader (*.ds.gba)")
        self._filter_gba_ds.add_pattern("*.nds")

        self._filter_any = Gtk.FileFilter()
        self._filter_any.set_name("All files")
        self._filter_any.add_pattern("*")

        self._filter_ds = Gtk.FileFilter()
        self._filter_ds.set_name("DeSmuME binary (*.ds)")
        self._filter_ds.add_pattern("*.ds")

        self._filter_dsm = Gtk.FileFilter()
        self._filter_dsm.set_name("DeSmuME move file (*.dsm)")
        self._filter_dsm.add_pattern("*.dsm")

        self._main_draw = builder.get_object("wDraw_Main")
        self._main_draw.set_events(Gdk.EventMask.ALL_EVENTS_MASK)
        self._main_draw.show()
        self._sub_draw = builder.get_object("wDraw_Sub")
        self._sub_draw.set_events(Gdk.EventMask.ALL_EVENTS_MASK)
        self._sub_draw.show()

        w_status: Gtk.Statusbar = self.builder.get_object("wStatus")
        self._movie_statusbar_context = w_status.get_context_id("Movie")

        builder.connect_signals(self)

    def on_destroy(self, *args):
        self.emu.destroy()
        Gtk.main_quit()

    def gtk_main_quit(self, *args):
        self.emu.destroy()
        Gtk.main_quit()

    def gtk_widget_hide_on_delete(self, w: Gtk.Widget, *args):
        w.hide_on_delete()
        return True

    def gtk_widget_hide(self, w: Gtk.Widget, *args):
        w.hide()

    # INPUT BUTTONS / KEYBOARD
    def on_wMainW_key_press_event(self, widget: Gtk.Widget, event: Gdk.EventKey, *args):
        key = self.lookup_key(event.keyval)
        # shift,ctrl, both alts
        mask = Gdk.ModifierType.SHIFT_MASK | Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.MOD1_MASK | Gdk.ModifierType.MOD5_MASK
        if event.state & mask == 0:
            if event.keyval == self._keyboard_cfg[Keys.KEY_BOOST - 1]:
                self._boost = not self._boost
                if self._boost:
                    self._save_fs = self._frameskip
                    self._frameskip = self._boost_fs
                else:
                    self._frameskip = self._save_fs
                return True
            elif key and self.emu.is_running():
                self.emu.input.keypad_add_key(key)
                return True
        return False

    def on_wMainW_key_release_event(self, widget: Gtk.Widget, event: Gdk.EventKey, *args):
        key = self.lookup_key(event.keyval)
        if key and self.emu.is_running():
            self.emu.input.keypad_rm_key(key)

    # OUTPUT SCREENS
    def on_wDrawScreen_main_draw_event(self, widget: Gtk.DrawingArea, ctx: cairo.Context, *args):
        return self.renderer.screen(self._current_screen_width, self._current_screen_height, ctx, 0 if not self._screen_invert else 1)

    def on_wDrawScreen_main_configure_event(self, widget: Gtk.DrawingArea, *args):
        self.renderer.reshape(widget, 0)
        return True

    def on_wDrawScreen_sub_draw_event(self, widget: Gtk.DrawingArea, ctx: cairo.Context, *args):
        return self.renderer.screen(self._current_screen_width, self._current_screen_height, ctx, 1 if not self._screen_invert else 0)

    def on_wDrawScreen_sub_configure_event(self, widget: Gtk.DrawingArea, *args):
        self.renderer.reshape(widget, 1)
        return True

    # INPUT STYLUS / MOUSE
    def on_wDrawScreen_main_motion_notify_event(self, widget: Gtk.Widget, event: Gdk.EventMotion, *args):
        return self.on_wDrawScreen_motion_notify_event(widget, event, 0)

    def on_wDrawScreen_main_button_release_event(self, widget: Gtk.Widget, event: Gdk.EventButton, *args):
        return self.on_wDrawScreen_button_release_event(widget, event, 0)

    def on_wDrawScreen_main_button_press_event(self, widget: Gtk.Widget, event: Gdk.EventButton, *args):
        return self.on_wDrawScreen_button_press_event(widget, event, 0)

    def on_wDrawScreen_sub_motion_notify_event(self, widget: Gtk.Widget, event: Gdk.EventMotion, *args):
        return self.on_wDrawScreen_motion_notify_event(widget, event, 1)

    def on_wDrawScreen_sub_button_release_event(self, widget: Gtk.Widget, event: Gdk.EventButton, *args):
        return self.on_wDrawScreen_button_release_event(widget, event, 1)

    def on_wDrawScreen_sub_button_press_event(self, widget: Gtk.Widget, event: Gdk.EventButton, *args):
        return self.on_wDrawScreen_button_press_event(widget, event, 1)

    def on_wDrawScreen_motion_notify_event(self, widget: Gtk.Widget, event: Gdk.EventMotion, display_id: int):
        if (display_id == 1 ^ self._screen_invert) and self._click:
            if event.is_hint:
                _, x, y, state = widget.get_window().get_pointer()
            else:
                x = event.x
                y = event.y
                state = event.state
            if state & Gdk.ModifierType.BUTTON1_MASK:
                self.set_touch_pos(x, y)

    def on_wDrawScreen_button_release_event(self, widget: Gtk.Widget, event: Gdk.EventButton, display_id: int):
        if display_id == 1 ^ self._screen_invert and self._click:
            self._click = False
            self.emu.input.touch_release()
        return True

    def on_wDrawScreen_button_press_event(self, widget: Gtk.Widget, event: Gdk.EventButton, display_id: int):
        if event.button == 1:
            if (display_id == 1 ^ self._screen_invert) and self.emu.is_running():
                self._click = True
                _, x, y, state = widget.get_window().get_pointer()
                if state & Gdk.ModifierType.BUTTON1_MASK:
                    self.set_touch_pos(x, y)
        elif event.button == 2:
            # filter out 2x / 3x clicks
            if event.type == Gdk.EventType.BUTTON_PRESS:
                self.rotate(self.renderer.get_screen_rotation() + 90)
        return True

    def on_wDrawScreen_scroll_event(self, widget: Gtk.Widget, event: Gdk.Event, *args):
        zoom_inc = .125
        zoom_min = 0.5
        zoom_max = 2.0

        scale = self.renderer.get_scale()
        if event.delta_y < 0:
            scale += zoom_inc
        elif event.delta_y > 0:
            scale -= zoom_inc

        if scale > zoom_max:
            scale = zoom_max
        elif scale < zoom_min:
            scale = zoom_min

        self.renderer.set_scale(scale)

        self.resize()

        return True

    # MENU FILE
    def on_menu_open_activate(self, *args):
        self.emu.pause()

        response, fn = self._file_chooser(Gtk.FileChooserAction.OPEN, "Open...", (self._filter_nds, self._filter_gba_ds, self._filter_any))

        if response == Gtk.ResponseType.OK:
            try:
                self.emu.open(fn)
            except RuntimeError:
                md = Gtk.MessageDialog(None,
                                       Gtk.DialogFlags.DESTROY_WITH_PARENT, Gtk.MessageType.ERROR,
                                       Gtk.ButtonsType.OK, f"Unable to load: {fn}",
                                       title="Error!")
                md.set_position(Gtk.WindowPosition.CENTER)
                md.run()
                md.destroy()
            else:
                self.resume()
                self.enable_rom_features()

    def on_menu_pscreen_activate(self, menu_item: Gtk.MenuItem, *args):
        was_running = self.emu.is_running()
        if was_running:
            self.emu.pause()

        filter_png = Gtk.FileFilter()
        filter_png.set_name("PNG Image (*.png)")
        filter_png.add_pattern("*.png")

        response, fn = self._file_chooser(Gtk.FileChooserAction.SAVE, "Save Screenshot...", (filter_png, self._filter_any))

        if response == Gtk.ResponseType.OK:
            self.emu.screenshot().save(fn)

        if was_running:
            self.resume()

    def on_menu_quit_activate(self, menu_item: Gtk.MenuItem, *args):
        self.gtk_main_quit()

    # MENU SAVES
    def on_loadstateXX_activate(self, w: Gtk.Widget, *args):
        slot = widget_to_primitive(w)
        self.emu.savestate.load(slot)

    def on_savestateXX_activate(self, w: Gtk.Widget, *args):
        slot = widget_to_primitive(w)
        self.emu.savestate.save(slot)
        self.update_savestate_menu("loadstate", slot)
        self.update_savestate_menu("savestate", slot)

    def on_loadstate_file_activate(self, *args):
        self.emu.pause()

        response, fn = self._file_chooser(Gtk.FileChooserAction.OPEN, "Load savestate...", (self._filter_ds, self._filter_any))

        if response == Gtk.ResponseType.OK:
            try:
                self.emu.savestate.load_file(fn)
            except RuntimeError as err:
                md = Gtk.MessageDialog(None,
                                       Gtk.DialogFlags.DESTROY_WITH_PARENT, Gtk.MessageType.ERROR,
                                       Gtk.ButtonsType.OK, str(err),
                                       title="Error!")
                md.set_position(Gtk.WindowPosition.CENTER)
                md.run()
                md.destroy()
            self.resume()

    def on_savestate_file_activate(self, *args):
        self.emu.pause()

        response, fn = self._file_chooser(Gtk.FileChooserAction.SAVE, "Save savestate...",  (self._filter_ds, self._filter_any))

        if response == Gtk.ResponseType.OK:
            try:
                self.emu.savestate.save_file(fn)
            except RuntimeError as err:
                md = Gtk.MessageDialog(None,
                                       Gtk.DialogFlags.DESTROY_WITH_PARENT, Gtk.MessageType.ERROR,
                                       Gtk.ButtonsType.OK, str(err),
                                       title="Error!")
                md.set_position(Gtk.WindowPosition.CENTER)
                md.run()
                md.destroy()
            self.resume()

    def on_savetypeXX_toggled(self, w: Gtk.Widget, *args):
        type = widget_to_primitive(w)
        self.emu.set_savetype(type)

    # MOVIE

    def on_menu_movie_play_activate(self, *args):
        self.emu.pause()

        response, fn = self._file_chooser(Gtk.FileChooserAction.OPEN, "Choose movie...", (self._filter_dsm, self._filter_any))

        if response == Gtk.ResponseType.OK:
            try:
                self.emu.movie.play(fn)
                self.push_statusbar_text(self._movie_statusbar_context, f"Playing movie: {fn}...")
            except RuntimeError as err:
                md = Gtk.MessageDialog(None,
                                       Gtk.DialogFlags.DESTROY_WITH_PARENT, Gtk.MessageType.ERROR,
                                       Gtk.ButtonsType.OK, str(err),
                                       title="Error!")
                md.set_position(Gtk.WindowPosition.CENTER)
                md.run()
                md.destroy()
            self.resume()

    def on_menu_movie_record_activate(self, *args):
        self.emu.pause()

        response, fn = self._file_chooser(Gtk.FileChooserAction.SAVE, "Save movie as...",  (self._filter_dsm, self._filter_any))

        if response == Gtk.ResponseType.OK:
            try:
                self.emu.movie.record(fn, "")
                self.push_statusbar_text(self._movie_statusbar_context, f"Recording movie: {fn}...")
            except RuntimeError as err:
                md = Gtk.MessageDialog(None,
                                       Gtk.DialogFlags.DESTROY_WITH_PARENT, Gtk.MessageType.ERROR,
                                       Gtk.ButtonsType.OK, str(err),
                                       title="Error!")
                md.set_position(Gtk.WindowPosition.CENTER)
                md.run()
                md.destroy()
            self.resume()

    def on_menu_movie_stop_activate(self, *args):
        self.emu.movie.stop()
        # TODO: We don't actually pop it when doing anything but pressing this button...
        self.pop_statusbar_text(self._movie_statusbar_context)

    # MENU EMULATION
    def on_menu_exec_activate(self, menu_item: Gtk.MenuItem, *args):
        self.resume()

    def on_menu_pause_activate(self, menu_item: Gtk.MenuItem, *args):
        self._supress_event = True
        self.builder.get_object("wgt_Exec").set_active(False)
        self._supress_event = False
        self.emu.pause()

    def on_menu_reset_activate(self, menu_item: Gtk.MenuItem, *args):
        self.emu.reset()

    def on_menu_layers_toggled(self, menu_item: Gtk.MenuItem, *args):
        w1 = self.builder.get_object("wvb_1_Main")
        w2 = self.builder.get_object("wvb_2_Sub")
        if menu_item.get_active():
            w1.show()
            w2.show()
        else:
            w1.hide()
            w2.hide()
        self._mainwindow_resize()

    def on_fsXX_activate(self, w, *args):
        self._frameskip = widget_to_primitive(w)

    def on_sizeXX_activate(self, w, *args):
        self.renderer.set_scale(widget_to_primitive(w))
        self.resize()

    # MENU CONFIG
    def on_menu_controls_activate(self, menu_item: Gtk.MenuItem, *args):
        new_keyboard_cfg = KeyboardControlsDialogController(self.window).run(self._keyboard_cfg)
        if new_keyboard_cfg is not None:
            self._keyboard_cfg = new_keyboard_cfg

    def on_menu_joy_controls_activate(self, menu_item: Gtk.MenuItem, *args):
        self._joystick_cfg = JoystickControlsDialogController(self.window).run(
            self._joystick_cfg, self.emu.input, self.emu.is_running()
        )

    def on_menu_audio_on_toggled(self, menu_item: Gtk.MenuItem, *args):
        if menu_item.get_active():
            self.emu.volume_set(100)
        else:
            self.emu.volume_set(0)

    def on_menu_gapscreen_toggled(self, menu_item: Gtk.MenuItem, *args):
        self._screen_gap = menu_item.get_active()
        self._mainwindow_resize()

    def on_menu_nogap_toggled(self, menu_item: Gtk.MenuItem, *args):
        self._screen_no_gap = menu_item.get_active()
        self._mainwindow_resize()

    def on_menu_rightscreen_toggled(self, menu_item: Gtk.MenuItem, *args):
        self.rightscreen(menu_item.get_active())

    def on_menu_rotatescreen_toggled(self, w, *args):
        self.rotate(widget_to_primitive(w))

    # MENU ?
    def on_menu_apropos_activate(self, menu_item: Gtk.MenuItem, *args):
        w_about: Gtk.AboutDialog = self.builder.get_object("wAboutDlg")
        w_about.run()

    # TOOLBAR
    def on_wgt_Open_clicked(self, btn: Gtk.ToolButton, *args):
        self.on_menu_open_activate()

    def on_wgt_Exec_toggled(self, btn: Gtk.ToggleToolButton, *args):
        if self._supress_event:
            return
        if btn.get_active():
            self.resume()
        else:
            self.emu.pause()

    def on_wgt_Reset_clicked(self, btn: Gtk.ToolButton, *args):
        self.emu.reset()

    # LAYERS TOGGLE
    def on_wc_1_BGXX_toggled(self, w, *args):
        slot = widget_to_primitive(w)
        self.emu.gpu_set_layer_main_enable_state(slot, self.builder.get_object(f"wc_1_BG{slot}").get_active())

    def on_wc_2_BGXX_toggled(self, w, *args):
        slot = widget_to_primitive(w)
        self.emu.gpu_set_layer_sub_enable_state(slot, self.builder.get_object(f"wc_2_BG{slot}").get_active())

    def lookup_key(self, keyval):
        key = False
        for i in range(0, Keys.NB_KEYS):
            if keyval == self._keyboard_cfg[i]:
                key = keymask(i)
                break
        return key

    def resume(self):
        self.emu.resume()
        self._supress_event = True
        self.builder.get_object("wgt_Exec").set_active(True)
        self._supress_event = False
        if not self._registered_main_loop:
            self._registered_main_loop = True
            GLib.idle_add(self.emu_cycle)

    def emu_cycle(self):
        if self.emu.is_running():
            skip_frame = False
            if self._frameskip != 0 and (self._fs_frame_count % (self._frameskip + 1) != 0):
                self.emu.skip_next_frame()
                skip_frame = True

            self._fs_frame_count += 1
            self._fps_frame_count += 1

            if not self._fps_sec_start:
                self._fps_sec_start = self.emu.get_ticks()
            if self.emu.get_ticks() - self._fps_sec_start >= 1000:
                self._fps_sec_start = self.emu.get_ticks()
                self._fps = self._fps_frame_count
                self._fps_frame_count = 0
                self.window.set_title(f"PyDeSmuME - {self._fps}fps")

            self.emu.cycle()

            if not skip_frame:
                self._main_draw.queue_draw()
                self._sub_draw.queue_draw()

            self._ticks_cur_frame = self.emu.get_ticks()

            if not self._fps_limiter_disabled:
                if self._ticks_cur_frame - self._ticks_prev_frame < TICKS_PER_FRAME:
                    while self._ticks_cur_frame - self._ticks_prev_frame < TICKS_PER_FRAME:
                        self._ticks_cur_frame = self.emu.get_ticks()

            self._ticks_prev_frame = self.emu.get_ticks()
            return True

        self._main_draw.queue_draw()
        self._sub_draw.queue_draw()
        self._registered_main_loop = False
        return False

    def enable_rom_features(self):
        self.emu.savestate.scan()
        self.update_savestates_menu()
        self._set_sensitve("menu_exec", True)
        self._set_sensitve("menu_pause", True)
        self._set_sensitve("menu_reset", True)
        self._set_sensitve("wgt_Exec", True)
        self._set_sensitve("wgt_Reset", True)
        self._set_sensitve("menu_movie_play", True)
        self._set_sensitve("menu_movie_record", True)
        self._set_sensitve("menu_movie_stop", True)

    def update_savestates_menu(self):
        for i in range(1, NB_STATES + 1):
            if self.emu.savestate.exists(i):
                self.update_savestate_menu("loadstate", i)
                self.update_savestate_menu("savestate", i)
            else:
                self.clear_savestate_menu("loadstate", i)
                self.clear_savestate_menu("savestate", i)

    def clear_savestate_menu(self, name, num):
        w: Gtk.MenuItem = self.builder.get_object(f"{name}{num}")
        w.get_child().set_text(f"State {num} (empty)")

    def update_savestate_menu(self, name, num):
        w = self.builder.get_object(f"{name}{num}")
        w.get_child().set_text(f"State {num} ({self.emu.savestate.date(num)})")

    def set_touch_pos(self, x: int, y: int):
        scale = self.renderer.get_scale()
        rotation = self.renderer.get_screen_rotation()
        x /= scale
        y /= scale
        emu_x = x
        emu_y = y
        if rotation == 90 or rotation == 270:
            emu_x = 256 -y
            emu_y = x

        if emu_x < 0:
            emu_x = 0
        elif emu_x > SCREEN_WIDTH - 1:
            emu_x = SCREEN_WIDTH - 1

        if emu_y < 9:
            emu_y = 0
        elif emu_y > SCREEN_HEIGHT:
            emu_y = SCREEN_HEIGHT

        if self._screen_invert:
            emu_x = SCREEN_WIDTH -1 - emu_x
            emu_y = SCREEN_HEIGHT - emu_y

        self.emu.input.touch_set_pos(int(emu_x), int(emu_y))

    def rotate(self, angle):
        if angle >= 360:
            angle -= 360
        self.renderer.set_screen_rotation(angle)
        rotated = angle == 90 or angle == 270
        self._screen_invert = angle >= 180
        self._current_screen_height = SCREEN_HEIGHT
        self._current_screen_width = SCREEN_WIDTH
        if rotated:
            self._current_screen_height = SCREEN_WIDTH
            self._current_screen_width = SCREEN_HEIGHT
        self.rightscreen(rotated)

        self.resize()

    def rightscreen(self, apply):
        table: Gtk.Table = self.builder.get_object("table_layout")
        chk = self.builder.get_object("wvb_2_Sub")
        self._screen_right = apply or self._screen_right_force
        if self._screen_right:
            gtk_table_reattach(table, self._sub_draw, 3, 4, 0, 1, 0, 0, 0, 0)
            gtk_table_reattach(table, chk, 4, 5, 0, 1, 0, 0, 0, 0)
        else:
            gtk_table_reattach(table, self._sub_draw, 1, 2, 2, 3, 0, 0, 0, 0)
            gtk_table_reattach(table, chk, 0, 1, 2, 3, 0, 0, 0, 0)

        table.queue_resize()
        self._mainwindow_resize()

    def resize(self):
        scale = self.renderer.get_scale()

        self._main_draw.set_size_request(self._current_screen_width * scale, self._current_screen_height * scale)
        self._sub_draw.set_size_request(self._current_screen_width * scale, self._current_screen_height * scale)

        self._mainwindow_resize()

    def _set_sensitve(self, name, state):
        w = self.builder.get_object(name)
        w.set_sensitive(state)

    def _mainwindow_resize(self):
        spacer1: Gtk.Widget = self.builder.get_object("misc_sep3")
        spacer2: Gtk.Widget = self.builder.get_object("misc_sep4")

        current_rotation = self.renderer.get_screen_rotation()
        rotated = current_rotation == 90 or current_rotation == 270

        dim1 = dim2 = 66 * self.renderer.get_scale()
        if not self._screen_gap:
            dim1 = dim2 = -1

        if self._screen_no_gap:
            spacer1.hide()
            spacer2.hide()
        else:
            spacer1.show()
            spacer2.show()

        if self._screen_right == rotated:
            if self._screen_right:
                dim2 = -1
            else:
                dim1 = -1

        spacer1.set_size_request(dim1, -1)
        spacer2.set_size_request(-1, dim2)
        self.window.resize(1, 1)

    def _file_chooser(self, type, name, filter):
        btn = Gtk.STOCK_OPEN
        if type == Gtk.FileChooserAction.SAVE:
            btn = Gtk.STOCK_SAVE
        dialog = Gtk.FileChooserDialog(
            name,
            self.window,
            type,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, btn, Gtk.ResponseType.OK)
        )
        for f in filter:
            dialog.add_filter(f)

        response = dialog.run()
        fn = dialog.get_filename()
        dialog.destroy()

        return response, fn

    def push_statusbar_text(self, context, text):
        w_status: Gtk.Statusbar = self.builder.get_object("wStatus")
        w_status.push(context, text)

    def pop_statusbar_text(self, context):
        w_status: Gtk.Statusbar = self.builder.get_object("wStatus")
        w_status.pop(context)


def gtk_table_reattach(
        table: Gtk.Table, w: Gtk.Widget, left_attach: int, right_attach: int, top_attach: int, bottom_attach: int,
        xoptions: int, yoptions: int, xpadding: int, ypadding: int):
    for child in table.get_children():
        if child == w:
            table.child_set_property(child, 'left_attach', left_attach)
            table.child_set_property(child, 'right_attach', right_attach)
            table.child_set_property(child, 'top_attach', top_attach)
            table.child_set_property(child, 'bottom_attach', bottom_attach)
            table.child_set_property(child, 'x-options', xoptions)
            table.child_set_property(child, 'x-padding', xpadding)
            table.child_set_property(child, 'y-options', yoptions)
            table.child_set_property(child, 'y-padding', ypadding)



#  Copyright 2020 Marco Köpcke (Parakoopa)
#
#  This file is part of py-desmume.
#
#  py-desmume is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  py-desmume is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with py-desmume.  If not, see <https://www.gnu.org/licenses/>.
__version__ = '0.0.4.post2'
from setuptools import setup, find_packages

from setuptools.command.build_ext import build_ext
from setuptools.command.develop import develop
from setuptools.command.install import install

import os
import platform
import shutil
import subprocess
import sys
from distutils.dist import Distribution
from glob import glob

# README read-in
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()
# END README read-in

is_installed_develop = False


class BinaryDistribution(Distribution):
    def is_pure(self):
        return False

    def has_ext_modules(self):
        return True


# Bug in distutils; see https://github.com/google/or-tools/issues/616#issuecomment-371480314
class InstallPlatlib(install):
    def finalize_options(self):
        install.finalize_options(self)
        if self.distribution.has_ext_modules():
            self.install_lib = self.install_platlib


class Develop(develop):
    def run(self):
        global is_installed_develop
        is_installed_develop = True
        super().run()


class BuildExt(build_ext):
    """Compiles the shared object using automake or the DLL using Visual Studio, depending on platform."""
    def run(self):
        # Don't build in develop mode
        if is_installed_develop:
            return

        this_path = os.getcwd()
        path_repo = os.path.join(this_path, 'desmume_src')
        path_interface = os.path.join(path_repo, 'desmume', 'src', 'frontend', 'interface')

        is_windows = platform.system() == "Windows"

        # Run the build script depending on the platform
        if is_windows:
            libraries = self.build_windows(path_interface)
        else:
            libraries = self.build_unix(path_interface)
        if not libraries:
            print("Could not compile the DeSmuME library.")
            print("")
            sys.exit(1)
        os.chdir(this_path)

        # Copy the libraries to the correct place
        for library in libraries:
            build_target = os.path.join(
                self.build_lib, 'desmume',
                os.path.basename(library)
            )
            print(f"Copying {library} -> {build_target}")
            shutil.copyfile(library, build_target)

    def build_unix(self, interface_path):
        """Tested against manylinux2014, see Jenkinsfile for requirements."""
        os.chdir(interface_path)
        environ = os.environ.copy()

        if platform.system() == "Darwin":
            environ["CC"] = "clang"
            environ["CXX"] = "clang++"

        print(f"BUILDING UNIX - meson build")
        retcode = subprocess.call(["meson", "build", "--buildtype=release"], env=environ)
        if retcode:
            return False
        print(f"BUILDING UNIX - ninja")
        retcode = subprocess.call(["ninja", "-C", "build"], env=environ)
        if retcode:
            return False

        if platform.system() == "Darwin":
            return [os.path.abspath(os.path.join(interface_path, 'build', 'libdesmume.dylib'))]
        else:
            return [os.path.abspath(os.path.join(interface_path, 'build', 'libdesmume.so'))]

    def build_windows(self, interface_path):
        """Requires Visual Studio."""
        is_64bits = sys.maxsize > 2 ** 32
        if os.getenv('BUILD_32', False):
            is_64bits = False
        arch_dirname = 'x64' if is_64bits else 'x86'
        arch_targetname = 'x64' if is_64bits else 'Win32'

        win_path = os.path.join(interface_path, "windows")
        os.chdir(win_path)
        print(f"BUILDING WINDOWS - msbuild.exe")
        config = "Release"
        if os.environ.get('FASTBUILD', False):
            config = "Release Fastbuild"
        retcode = subprocess.call(["MSBuild.exe", "DeSmuME_Interface.vcxproj", f"/p:configuration={config}", f"/p:Platform={arch_targetname}"])
        if retcode:
            return False

        # Find and rename DeSmuME DLL file
        path_dll = glob(os.path.join(win_path, '__bins', '*.dll'))[0]
        new_name_dll = os.path.join(os.path.dirname(path_dll), 'libdesmume.dll')
        os.rename(path_dll, new_name_dll)

        return [
            os.path.abspath(new_name_dll),
            # Also include the required SDL lib
            os.path.abspath(os.path.join(win_path, 'SDL', 'lib', arch_dirname, 'SDL2.dll'))
        ]

setup(
    name='py-desmume',
    version=__version__,
    packages=find_packages(),
    package_data={'desmume': ['**/*.css', '**/*.glade', '**/control_ui/*.glade', 'libdesmume.so', 'libdesmume.dylib', '*.dll']},
    description='Python library to interface with DeSmuME, the Nintendo DS emulator + sample GTK-based frontend',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/SkyTemple/py-desmume/',
    install_requires=[
        'Pillow >= 6.1.0'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    distclass=BinaryDistribution,
    cmdclass={'build_ext': BuildExt, 'install': InstallPlatlib, 'develop': Develop}
)



"""
This module contains the Python interface for DeSmuME.

:class:`DeSmuME` is the main entrypoint to load and interact with the emulator.
"""
#  Copyright 2020 Marco Köpcke (Parakoopa)
#
#  This file is part of py-desmume.
#
#  py-desmume is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  py-desmume is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with py-desmume.  If not, see <https://www.gnu.org/licenses/>.
import os
import platform
from ctypes import cdll, create_string_buffer, cast, c_char_p, POINTER, c_int, c_char, c_uint16, c_uint8, Structure, \
    c_uint, CFUNCTYPE, c_int8, c_int16, c_uint32, c_int32
from enum import Enum
from typing import Union, Callable, List, Optional

try:
    from PIL import Image
except ImportError:
    from pil import Image

from desmume.controls import add_key, rm_key


SCREEN_WIDTH = 256
SCREEN_HEIGHT = 192
SCREEN_HEIGHT_BOTH = SCREEN_HEIGHT * 2
SCREEN_PIXEL_SIZE = SCREEN_WIDTH * SCREEN_HEIGHT
SCREEN_PIXEL_SIZE_BOTH = SCREEN_WIDTH * SCREEN_HEIGHT_BOTH
NB_STATES = 10

MEMORY_CB_FN = CFUNCTYPE(None, c_uint, c_int)

MemoryCbFn = Callable[[int, int], None]
"""
A callback function for the memory hooks. The first parameter of the callback function will receive the watched address,
the second the size of the event that triggered it, in bytes (eg. how many bytes were written / read).
"""


def strbytes(s):
    return s.encode('utf-8')


class Language(Enum):
    """Language codes."""
    JAPANESE = 0
    ENGLISH = 1
    FRENCH = 2
    GERMAN = 3
    ITALIAN = 4
    SPANISH = 5


class StartFrom(Enum):
    """Start codes for movie recording."""
    START_BLANK = 0
    START_SRAM = 1
    START_SAVESTATE = 2


class DeSmuME_SDL_Window:
    """
    A window that displays the emulator and processes touchscreen and keyboard inputs (default keyboard
    configuration only).
    This is meant to be a simple way to use and test the library, for intergration in custom UIs you
    probably want to process input and display manually.

    Should not be instantiated manually!
    """
    def __init__(self, emu: 'DeSmuME', auto_pause=True, use_opengl_if_possible=True):
        self.lib = emu.lib
        self.lib.desmume_draw_window_init(bool(auto_pause), bool(use_opengl_if_possible))

    def __del__(self):
        self.destroy()

    def destroy(self):
        """Destroy the window."""
        self.lib.desmume_draw_window_free()

    def draw(self):
        """Draw the current framebuffer to the window."""
        self.lib.desmume_draw_window_frame()

    def process_input(self):
        """Process the touchscreen input for the current cycle."""
        self.lib.desmume_draw_window_input()

    def has_quit(self) -> bool:
        """Returns true, when the window was closed by the user."""
        return bool(self.lib.desmume_draw_window_has_quit())


class DeSmuME_Input:
    """Manage input processing for the emulator. Should not be instantiated manually!"""
    def __init__(self, emu: 'DeSmuME'):
        self.emu = emu
        self.lib = emu.lib
        self.has_joy = False

    def __del__(self):
        if self.has_joy:
            self.joy_uninit()

    def joy_init(self):
        """Initialize the joystick input processing. Call this to enable automatic joystick input processing."""
        if not self.has_joy:
            self.emu.lib.desmume_input_joy_init()
            self.has_joy = True

    def joy_uninit(self):
        """De-initialize the joystick input processing."""
        if self.has_joy:
            self.lib.desmume_input_joy_uninit()

    def joy_number_connected(self) -> int:
        """Returns the number of connected joysticks. Joysticks must be initialized."""
        if self.has_joy:
            return self.emu.lib.desmume_input_joy_number_connected()
        raise ValueError("Joystick not initialized.")

    def joy_get_key(self, index: int) -> int:
        """Get the joystick key assigned to the specified emulator key. Joysticks must be initialized."""
        if self.has_joy:
            return self.emu.lib.desmume_input_joy_get_key(index)
        raise ValueError("Joystick not initialized.")

    def joy_get_set_key(self, index: int) -> int:
        """
        Pause the thread and wait for the user to press a button.
        This button will be assigned to the specified emulator key. Joysticks must be initialized.
        """
        if self.has_joy:
            return self.emu.lib.desmume_input_joy_get_set_key(index)
        raise ValueError("Joystick not initialized.")

    def joy_set_key(self, index: int, joystick_key_index: int):
        """
        Sets the emulator key ``index`` to the specified joystick key ``joystick_key_index``.
        Joysticks must be initialized.
        """
        if self.has_joy:
            self.emu.lib.desmume_input_joy_set_key(index, joystick_key_index)
            return
        raise ValueError("Joystick not initialized.")

    def keypad_update(self, keys: int) -> int:
        """
        Update the keypad (pressed DS buttons) of currently pressed emulator keys.
        You should probably use ``keypad_add_key`` and ``keypad_rm_key`` instead.
        """
        self.emu.lib.desmume_input_keypad_update.argtypes = [c_uint16]
        return self.emu.lib.desmume_input_keypad_update(keys)

    def keypad_get(self) -> int:
        """Returns the current emulator key keypad (pressed DS buttons)."""
        self.emu.lib.desmume_input_keypad_update.restype = c_uint16
        return self.emu.lib.desmume_input_keypad_get()

    def keypad_add_key(self, key: int):
        """
        Adds a key to the emulators current keymask (presses it). To be used with ``keymask``:

        >>> from desmume.controls import keymask, Keys
        >>> keym = keymask(Keys.KEY_A)
        >>> DeSmuME().input.keypad_add_key(keym)
        """
        old_keypad = self.keypad_get()
        self.keypad_update(add_key(old_keypad, key))

    def keypad_rm_key(self, key: int):
        """
        Removes a key from the emulators current keymask (releases it).
        See ``keypad_add_key`` for a usage example.
        """
        old_keypad = self.keypad_get()
        self.keypad_update(rm_key(old_keypad, key))

    def touch_set_pos(self, x: int, y: int):
        """Set the specified coordinate of the screen to be touched."""
        self.emu.lib.desmume_input_set_touch_pos(x, y)

    def touch_release(self):
        """Tell the emulator, that the user released touching the screen."""
        self.emu.lib.desmume_input_release_touch()


class DeSmuME_Savestate:
    """
    Load and save savestates. Either slots can be used  (maximum number of slots is in the constant ``NB_STATES``),
    or savestates can be directly loaded from / saved to files.

    Should not be instantiated manually!
    """
    def __init__(self, emu: 'DeSmuME'):
        self.emu = emu

    def scan(self):
        """Scan all savestate slots for if they exist or not. Required to be called before calling ``exists``."""
        self.emu.lib.desmume_savestate_scan()

    def exists(self, slot_id: int) -> bool:
        """Returns whether or not a savestate in the specified slot exists."""
        return bool(self.emu.lib.desmume_savestate_slot_exists(slot_id))

    def load(self, slot_id: int):
        """Load the savestate in the specified slot. It needs to exist."""
        return self.emu.lib.desmume_savestate_slot_load(slot_id)

    def save(self, slot_id: int):
        """Save the current game state to the savestate in the specified slot."""
        return self.emu.lib.desmume_savestate_slot_save(slot_id)

    def load_file(self, file_name: str):
        """
        Load a savestate from file.

        :raise: RuntimeError If the savestate could not be loaded.
        """
        if not self.emu.lib.desmume_savestate_load(c_char_p(strbytes(file_name))):
            raise RuntimeError("Unable to load savesate.")

    def save_file(self, file_name: str):
        """
        Save a savestate to file.

        :raise: RuntimeError If the savestate could not be saved.
        """
        if not self.emu.lib.desmume_savestate_save(c_char_p(strbytes(file_name))):
            raise RuntimeError("Unable to save savesate.")

    def date(self, slot_id: int) -> str:
        """Return the date a savestate was saved at as a string."""
        self.emu.lib.desmume_savestate_slot_date.restype = c_char_p
        return str(self.emu.lib.desmume_savestate_slot_date(slot_id), 'utf-8')


class DeSmuME_Date(Structure):
    """A date C struct, to be used with setting a date for movie recording."""
    _fields_ = [
        ("year", c_int),
        ("month", c_int),
        ("day", c_int),
        ("hour", c_int),
        ("minute", c_int),
        ("second", c_int),
        ("millisecond", c_int),
    ]


class DeSmuME_Movie:
    """Record and play movies. Should not be instantiated manually!"""
    def __init__(self, emu: 'DeSmuME'):
        self.emu = emu

    def play(self, file_name: str):
        """
        Load a movie file from a file and play it back.

        :raise: RuntimeError If playback failed.
        """
        self.emu.lib.desmume_movie_play.restype = c_char_p
        err = self.emu.lib.desmume_movie_play(c_char_p(strbytes(file_name)))
        if err is not None and err != "":
            raise RuntimeError(str(err, 'utf-8'))

    def record(
            self, file_name: str, author_name: str,
            start_from: StartFrom = StartFrom.START_BLANK, sram_save: str = "", rtc_date: DeSmuME_Date = None
    ):
        """
        Record a movie.

        :param file_name: The name of the file to save to.
        :param author_name: Name of the author of the movie.
        :param start_from: Where to start the recording from.
        :param sram_save: Filename of the SRAM save to use, optional.
        :param rtc_date: Date to set the real-time-clock to, optional, defaults to now.
        :return:
        """
        if not rtc_date:
            self.emu.lib.desmume_movie_record(
                c_char_p(strbytes(file_name)), c_char_p(strbytes(author_name)), start_from.value, c_char_p(strbytes(sram_save))
            )
        else:
            self.emu.lib.desmume_movie_record_from_date(
                c_char_p(strbytes(file_name)), c_char_p(strbytes(author_name)), start_from.value, c_char_p(strbytes(sram_save)), rtc_date
            )

    def stop(self):
        """Stops the current movie playback."""
        self.emu.lib.desmume_movie_stop()

    def is_active(self):
        return bool(self.emu.lib.desmume_movie_is_active())

    def is_recording(self):
        return bool(self.emu.lib.desmume_movie_is_recording())

    def is_playing(self):
        return bool(self.emu.lib.desmume_movie_is_playing())

    def is_finished(self):
        return bool(self.emu.lib.desmume_movie_is_finished())

    def get_length(self):
        if self.is_active():
            return self.emu.lib.desmume_movie_get_length()
        raise ValueError("No movie is active.")

    def get_name(self):
        if self.is_active():
            self.emu.lib.desmume_movie_get_name.restype = c_char_p
            return self.emu.lib.desmume_movie_get_name()
        raise ValueError("No movie is active.")

    def get_rerecord_count(self):
        if self.is_active():
            return self.emu.lib.desmume_movie_get_rerecord_count()
        raise ValueError("No movie is active.")

    def set_rerecord_count(self, count: int):
        if self.is_active():
            return self.emu.lib.desmume_movie_set_rerecord_count(count)
        raise ValueError("No movie is active.")

    def get_readonly(self):
        if self.is_active():
            return bool(self.emu.lib.desmume_movie_get_readonly())
        raise ValueError("No movie is active.")

    def set_readonly(self, state: bool):
        if self.is_active():
            return self.emu.lib.desmume_movie_set_readonly(state)
        raise ValueError("No movie is active.")


class MemoryAccessor:
    """
    Pythonic accessor class for manipulating the memory.

    Access a single unsigned byte of memory at offset 0x100:

    >>> emu_memory = DeSmuME().memory
    >>> emu_memory.unsigned[0x100]
    <<< 3

    Access a subset of the memory as bytes:

    >>> emu_memory = DeSmuME().memory
    >>> emu_memory.unsigned[0x100:0x200]
    <<< bytes(...)

    Access a subset of the memory as signed integers:

    >>> emu_memory = DeSmuME().memory
    >>> emu_memory.signed[0x100:0x200]
    <<< [-3, 2, ...]

    Access a subset of memory as 4-byte unsigned integers:

    >>> emu_memory = DeSmuME().memory
    >>> emu_memory.unsigned[0x100:0x200:4]
    <<< [-236, 1002, ...]

    Writing to memory works the same way. It doesn't matter if you use the signed or unsigned accessor for this, both
    work the same:

    >>> emu_memory = DeSmuME().memory
    >>> emu_memory.unsigned[0x100] = 2

    You can also use the ``read_*`` methods instead for a more verbose way to read the memory as integers.
    For writing those methods can be found in the ``DeSmuME_Memory`` parent object.

    Should not be instantiated manually!

    """
    def __init__(self, signed, mem: 'DeSmuME_Memory'):
        self.signed = signed
        self.mem = mem

    def __getitem__(self, key: Union[int, slice]) -> Union[int, bytes, List[int]]:
        if isinstance(key, int):
            return self.mem.read(key, key, 1, self.signed)
        return self.mem.read(key.start, key.stop, key.step, self.signed)

    def __setitem__(self, key: Union[int, slice], value: Union[int, bytes, List[int]]):
        if isinstance(key, int):
            return self.mem.write(key, key, 1, bytes([value]))
        return self.mem.write(key.start, key.stop, key.step, value)

    def read_byte(self, addr: int) -> int:
        """Read a 1-byte size integer at the specified address."""
        return self[addr]

    def read_short(self, addr: int) -> int:
        """Read a 2-byte size integer at the specified address."""
        return self[addr:addr:2]

    def read_long(self, addr: int) -> int:
        """Read a 2-byte size integer at the specified address."""
        return self[addr:addr:4]


class RegisterAccessor:
    """
    Access the registers of the emulator. The properties are the register names.
    ``rX`` are the numbered register names, some registers are available via aliases.
    Returned are integers.

    You can also get a register value by accessing this object:

    >>> DeSmuME().memory.register_arm9[0] == DeSmuME().memory.register_arm9.r0
    <<< True

    Should not be instantiated manually!

    """
    def __init__(self, prefix, mem: 'DeSmuME_Memory'):
        self.prefix = prefix
        self.lib = mem.emu.lib

    # <editor-fold desc="register accessors" defaultstate="collapsed">

    def __getitem__(self, item):
        if item == 0:
            return self.r0
        if item == 1:
            return self.r1
        if item == 2:
            return self.r2
        if item == 3:
            return self.r3
        if item == 4:
            return self.r4
        if item == 5:
            return self.r5
        if item == 6:
            return self.r6
        if item == 7:
            return self.r7
        if item == 8:
            return self.r8
        if item == 9:
            return self.r9
        if item == 10:
            return self.r10
        if item == 11:
            return self.r11
        if item == 12:
            return self.r12
        if item == 13:
            return self.r13
        if item == 14:
            return self.r14
        if item == 15:
            return self.r15
        raise ValueError("Invalid register")

    def __setitem__(self, item, value):
        if item == 0:
            self.r0 = value
        if item == 1:
            self.r1 = value
        if item == 2:
            self.r2 = value
        if item == 3:
            self.r3 = value
        if item == 4:
            self.r4 = value
        if item == 5:
            self.r5 = value
        if item == 6:
            self.r6 = value
        if item == 7:
            self.r7 = value
        if item == 8:
            self.r8 = value
        if item == 9:
            self.r9 = value
        if item == 10:
            self.r10 = value
        if item == 11:
            self.r11 = value
        if item == 12:
            self.r12 = value
        if item == 13:
            self.r13 = value
        if item == 14:
            self.r14 = value
        if item == 15:
            self.r15 = value
        raise ValueError("Invalid register")

    @property
    def r0(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r0")))

    @r0.setter
    def r0(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r0")), value)

    @property
    def r1(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r1")))

    @r1.setter
    def r1(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r1")), value)

    @property
    def r2(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r2")))

    @r2.setter
    def r2(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r2")), value)

    @property
    def r3(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r3")))

    @r3.setter
    def r3(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r3")), value)

    @property
    def r4(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r4")))

    @r4.setter
    def r4(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r4")), value)

    @property
    def r5(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r5")))

    @r5.setter
    def r5(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r5")), value)

    @property
    def r6(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r6")))

    @r6.setter
    def r6(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r6")), value)

    @property
    def r7(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r7")))

    @r7.setter
    def r7(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r7")), value)

    @property
    def r8(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r8")))

    @r8.setter
    def r8(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r8")), value)

    @property
    def r9(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r9")))

    @r9.setter
    def r9(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r9")), value)

    @property
    def r10(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r10")))

    @r10.setter
    def r10(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r10")), value)

    @property
    def r11(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r11")))

    @r11.setter
    def r11(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r11")), value)

    @property
    def r12(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r12")))

    @r12.setter
    def r12(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r12")), value)

    @property
    def r13(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r13")))

    @r13.setter
    def r13(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r13")), value)

    @property
    def r14(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r14")))

    @r14.setter
    def r14(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r14")), value)

    @property
    def r15(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "r15")))

    @r15.setter
    def r15(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "r15")), value)

    @property
    def cpsr(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "cpsr")))

    @cpsr.setter
    def cpsr(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "cpsr")), value)

    @property
    def spsr(self):
        return self.lib.desmume_memory_read_register(c_char_p(strbytes(self.prefix + "spsr")))

    @spsr.setter
    def spsr(self, value):
        self.lib.desmume_memory_write_register(c_char_p(strbytes(self.prefix + "spsr")), value)

    # Aliases
    @property
    def sp(self):
        """Alias for r13."""
        return self.r13

    @sp.setter
    def sp(self, value):
        self.r13 = value

    @property
    def lr(self):
        """Alias for r14."""
        return self.r14

    @lr.setter
    def lr(self, value):
        self.r14 = value

    @property
    def pc(self):
        """Alias for r15."""
        return self.r15

    @pc.setter
    def pc(self, value):
        self.r15 = value

    # </editor-fold>
RegisterAccesor = RegisterAccessor  # Typo alias for backwards-comaptibility.


class DeSmuME_Memory:
    """Access and manipulate the memory of the emulator. Should not be instantiated manually!"""
    def __init__(self, emu: 'DeSmuME'):
        self.emu = emu
        self._unsigned: MemoryAccessor = MemoryAccessor(False, self)
        self._signed: MemoryAccessor = MemoryAccessor(True, self)
        self._register_arm9: RegisterAccessor = RegisterAccessor("arm9.", self)
        self._register_arm7: RegisterAccessor = RegisterAccessor("arm7.", self)

        self.emu.lib.desmume_memory_read_byte.restype = c_uint8
        self.emu.lib.desmume_memory_read_byte_signed.restype = c_int8
        self.emu.lib.desmume_memory_read_short.restype = c_uint16
        self.emu.lib.desmume_memory_read_short_signed.restype = c_int16
        self.emu.lib.desmume_memory_read_long.restype = c_uint32
        self.emu.lib.desmume_memory_read_long_signed.restype = c_int32

        self.emu.lib.desmume_memory_write_byte.argtypes = [c_int, c_uint8]
        self.emu.lib.desmume_memory_write_short.argtypes = [c_int, c_uint16]
        self.emu.lib.desmume_memory_write_long.argtypes = [c_int, c_uint32]

        self.emu.lib.desmume_memory_read_register.argtypes = [c_char_p]
        self.emu.lib.desmume_memory_write_register.argtypes = [c_char_p, c_int]

        # Used to make sure that callbacks aren't GCed.
        # TODO: Some way to clean this up again.
        self._registered_cbs = []

    @property
    def unsigned(self) -> MemoryAccessor:
        """
        The accessor for accessing the memory values as raw unsigned bytes/ints.

        :type: MemoryAccessor
        """
        return self._unsigned

    @property
    def signed(self) -> MemoryAccessor:
        """
        The accessor for accessing the memory values as signed ints.

        :type: MemoryAccessor
        """
        return self._signed

    @property
    def register_arm9(self) -> RegisterAccessor:
        """
        ARM9 Registers.

        :type: RegisterAccessor
        """
        return self._register_arm9

    @property
    def register_arm7(self) -> RegisterAccessor:
        """
        ARM7 Registers.

        :type: RegisterAccessor
        """
        return self._register_arm7

    def read(self, start: int, end: int, size: int, signed: bool) -> Union[int, bytes, List[int]]:
        """
        Read part of NDS memory. You probably don't want to use this. Use the ``unsigned`` and ``signed``
        properties instead.

        Allowed sizes: 1 = byte, 2 = short, 4 = long

        If start and end are equal, returns an integer based on the size and the signed flag.

        If not, but size is 1 and signed is False, returns a bytes object with all bytes between start and end.
        Otherwise returns a list of ints based on size and signed.
        """
        if size is None:
            size = 1
        if start == end:
            # Read a single value, what kind depends on size
            if signed:
                if size == 1:
                    return self.emu.lib.desmume_memory_read_byte_signed(start)
                elif size == 2:
                    return self.emu.lib.desmume_memory_read_short_signed(start)
                elif size == 4:
                    return self.emu.lib.desmume_memory_read_long_signed(start)
            else:
                if size == 1:
                    return self.emu.lib.desmume_memory_read_byte(start)
                elif size == 2:
                    return self.emu.lib.desmume_memory_read_short(start)
                elif size == 4:
                    return self.emu.lib.desmume_memory_read_long(start)
            raise ValueError("Invalid size.")
        # Read a range.
        if signed:
            if size == 1:
                b = []
                for i, addr in enumerate(range(start, end, size)):
                    b.append(self.emu.lib.desmume_memory_read_byte_signed(addr))
            elif size == 2:
                b = []
                for i, addr in enumerate(range(start, end, size)):
                    b.append(self.emu.lib.desmume_memory_read_short_signed(addr))
            elif size == 4:
                b = []
                for i, addr in enumerate(range(start, end, size)):
                    b.append(self.emu.lib.desmume_memory_read_long_signed(addr))
            else:
                raise ValueError("Invalid size.")
        else:
            if size == 1:
                b = bytearray(int(end - start))
                for i, addr in enumerate(range(start, end, size)):
                    b[i] = self.emu.lib.desmume_memory_read_byte(addr)
                b = bytes(b)
            elif size == 2:
                b = []
                for i, addr in enumerate(range(start, end, size)):
                    b.append(self.emu.lib.desmume_memory_read_short(addr))
            elif size == 4:
                b = []
                for i, addr in enumerate(range(start, end, size)):
                    b.append(self.emu.lib.desmume_memory_read_long(addr))
            else:
                raise ValueError("Invalid size.")
        return b

    def write(self, start: int, end: int, size: int, value: Union[bytes, List[int]]):
        """
        Write part of NDS memory. You probably don't want to use this. Use the ``unsigned`` and ``signed``
        properties instead, or the ``write_*`` methods.
        """
        if start == end:
            end += 1  # Write at least one.
        if size == 1:
            for i, addr in enumerate(range(start, end, size)):
                self.emu.lib.desmume_memory_write_byte(addr, c_uint8(value[i]))
        elif size == 2:
            for i, addr in enumerate(range(start, end, size)):
                # value must be List[int]
                self.emu.lib.desmume_memory_write_short(addr, c_uint16(value[i]))
        elif size == 4:
            for i, addr in enumerate(range(start, end, size)):
                # value must be List[int]
                self.emu.lib.desmume_memory_write_long(addr, c_uint32(value[i]))
        else:
            raise ValueError("Invalid size.")

    def read_string(self, address: int, codec='windows-1255'):
        """Read a null-terminated string, beginning at address."""
        max_len = 50
        string_buff = bytearray(max_len)
        cur_byte = self.unsigned.read_byte(address)
        i = 0
        while cur_byte != 0:
            if i >= max_len:
                max_len += 50
                string_buff += bytearray(50)
            string_buff[i] = cur_byte
            i += 1
            cur_byte = self.unsigned.read_byte(address + i)
        return str(memoryview(string_buff)[:i], codec, 'ignore')

    def write_byte(self, addr: int, value: int):
        """Write a 1-byte integer to the memory at the specified address."""
        self.write(addr, addr, 1, [value])

    def write_short(self, addr: int, value: int):
        """Write a 2-byte integer to the memory at the specified address."""
        self.write(addr, addr, 2, [value])

    def write_long(self, addr: int, value: int):
        """Write a 4-byte integer to the memory at the specified address."""
        self.write(addr, addr, 4, [value])

    def get_next_instruction(self):
        """Returns the next instruction to be executed by the ARM9 processor."""
        self.emu.lib.desmume_memory_get_next_instruction()

    def set_next_instruction(self, address: int):
        """
        Sets the next instruction to be executed by the ARM9 processor.
        You should probably also consider updating the PC.
        """
        self.emu.lib.desmume_memory_set_next_instruction(address)

    def register_write(self, address: int, callback: Optional[MemoryCbFn], size=1):
        """
        Add a memory callback for when the memory at the specified address was changed.

        Setting a callback will override the previously registered one for this address.
        Set callback to None, to remove the callback for this address.

        A usage example can be found for ``register_exec``.

        :param address: The address to monitor.
        :param callback: This callback will be called when the memory was written to. See ``MemoryCbFn``.
        :param size: The maximum size that will be watched. If you set this to 4 for example,
                     a range of (address, address + 3) will be monitored.
        """
        casted_cbfn = callback
        if callback is not None:
            casted_cbfn = MEMORY_CB_FN(callback)
            self._registered_cbs.append(casted_cbfn)
        self.emu.lib.desmume_memory_register_write(address, size, casted_cbfn)

    def register_read(self, address: int, callback: Optional[MemoryCbFn], size=1):
        """
        Add a memory callback for when the memory at the specified address was read.

        Setting a callback will override the previously registered one for this address.
        Set callback to None, to remove the callback for this address.

        A usage example can be found for ``register_exec``.

        :param address: The address to monitor.
        :param callback: This callback will be called when the memory was read. See ``MemoryCbFn``.
        :param size: The maximum size that will be watched. If you set this to 4 for example,
                     a range of (address, address + 3) will be monitored.
        """
        casted_cbfn = callback
        if callback is not None:
            casted_cbfn = MEMORY_CB_FN(callback)
            self._registered_cbs.append(casted_cbfn)
        self.emu.lib.desmume_memory_register_read(address, size, casted_cbfn)

    def register_exec(self, address: int, callback: Optional[MemoryCbFn], size=2):
        """
        Add a memory callback for when the PC processed the operation at the specified address.

        Setting a callback will override the previously registered one for this address.
        Set callback to None, to remove the callback for this address.

        Example (will print 'Hello World', whenever the PC executes the code at 0x022f8818):

        >>> def my_callback(address, size):
        >>>     print("Hello World!")
        >>>
        >>> DeSmuME().memory.register_exec(0x022f8818, my_callback)

        :param address: The address to monitor.
        :param callback: This callback will be called when the operation was executed. See ``MemoryCbFn``.
        :param size: Leave this at 2.
        """
        casted_cbfn = callback
        if callback is not None:
            casted_cbfn = MEMORY_CB_FN(callback)
            self._registered_cbs.append(casted_cbfn)
        self.emu.lib.desmume_memory_register_exec(address, size, casted_cbfn)


class DeSmuME:
    """DeSmuME, the Nintendo DS emulator."""
    def __init__(self, dl_name: str = None):
        """
        Initializes a new emulator instance.
        A path to the shard library for DeSmuME can be passed, if not it will be auto-detected.
        """
        self.lib = None

        # Load the correct library
        if dl_name is None:
            # Try autodetect / CWD
            try:
                if platform.system().lower().startswith('windows'):
                    dl_name = "libdesmume.dll"
                    os.add_dll_directory(os.getcwd())
                elif platform.system().lower().startswith('linux'):
                    dl_name = "libdesmume.so"
                elif platform.system().lower().startswith('darwin'):
                    dl_name = "libdesmume.dylib"
                else:
                    RuntimeError(f"Unknown platform {platform.system()}, can't autodetect DLL to load.")

                self.lib = cdll.LoadLibrary(dl_name)
            except OSError:
                # Okay now try the package directory
                dl_name = os.path.dirname(os.path.realpath(__file__))
                if platform.system().lower().startswith('windows'):
                    os.add_dll_directory(dl_name)
                    dl_name = os.path.join(dl_name, "libdesmume.dll")
                elif platform.system().lower().startswith('linux'):
                    dl_name = os.path.join(dl_name, "libdesmume.so")
                elif platform.system().lower().startswith('darwin'):
                    dl_name = os.path.join(dl_name, "libdesmume.dylib")

                self.lib = cdll.LoadLibrary(dl_name)
        else:
            if platform.system().lower().startswith('windows'):
                os.add_dll_directory(os.path.dirname(dl_name))
                dl_name = os.path.basename(dl_name)

            self.lib = cdll.LoadLibrary(dl_name)

        self.lib.desmume_set_savetype(0)

        if self.lib.desmume_init() < 0:
            raise RuntimeError("Failed to init DeSmuME")

        self._input = DeSmuME_Input(self)
        self._savestate = DeSmuME_Savestate(self)
        self._movie = DeSmuME_Movie(self)
        self._memory = DeSmuME_Memory(self)
        self._sdl_window = None
        self._raw_buffer_rgbx = None

    def __del__(self):
        if self.lib is not None:
            self.lib.desmume_free()
            self._input.__del__()
            if self._sdl_window:
                self._sdl_window.__del__()
            self.lib = None

    def destroy(self):
        """Destroy the emulator and free memory. Calling any method after this WILL result in a crash."""
        self.__del__()

    @property
    def input(self):
        """
        Keyboard and joystick configuration.

        :type: DeSmuME_Input
        """
        return self._input

    @property
    def savestate(self):
        """
        Loading and saving savestates.

        :type: DeSmuME_Savestate
        """
        return self._savestate

    @property
    def movie(self):
        """
        Recording and playing back movies.

        :type: DeSmuME_Movie
        """
        return self._movie

    @property
    def memory(self):
        """
        Accessing and manipulating the memory.

        :type: DeSmuME_Memory
        """
        return self._memory

    def set_language(self, lang: Language):
        """Set the current firmware language."""
        self.lib.desmume_set_language(lang.value)

    def open(self, file_name: str, auto_resume=True):
        """
        Open a ROM by file name.

        If ``auto_resume`` is True, the emulator will automatically begin emulating the game.
        Otherwise the emulator is paused and you may call ``resume`` to unpause it.
        """
        if self.lib.desmume_open(c_char_p(strbytes(file_name))) < 0:
            raise RuntimeError("Unable to open ROM file.")
        if auto_resume:
            self.resume()

    def close(self):
        """
        Close a previously opened ROM, freeing up memory.
        You don't need to call this before opening a new ROM (it is done automatically).
        """
        self.lib.desmume_close()

    def set_savetype(self, value: int):
        """
        Set the type of the SRAM. Please see the DeSmuME documentation for possible values.
        0 is auto-detect and set by default.
        """
        # TODO: Enum?
        self.lib.desmume_set_savetype(value)

    def pause(self):
        """Pause the emulator."""
        self.lib.desmume_pause()

    def resume(self, keep_keypad=False):
        """
        Resume / unpause the emulator. This will reset the keypad (release all keys),
        except if ``keep_keypad`` is provided.
        """
        if keep_keypad:
            self.input.keypad_update(0)
        self.lib.desmume_resume()

    def reset(self):
        """Resets the emulator / restarts the current game."""
        self.lib.desmume_reset()

    def is_running(self) -> bool:
        """Returns ``True``, if a game is loaded and the emulator is running (not paused)."""
        return bool(self.lib.desmume_running())

    def skip_next_frame(self):
        """Tell the emulator to skip the next frame."""
        self.lib.desmume_skip_next_frame()

    def cycle(self, with_joystick=True):
        """
        Cycle one game cycle / frame. Set ``with_joystick`` to
        ``False``, if joystick processing was not initialized.
        """
        self.lib.desmume_cycle(with_joystick)

    def has_opengl(self) -> bool:
        """Returns ``True``, if OpenGL is available for rendering."""
        return bool(self.lib.desmume_has_opengl())

    def create_sdl_window(self, auto_pause=False, use_opengl_if_possible=True) -> DeSmuME_SDL_Window:
        """
        Create an SDL window for drawing the
        emulator in and processing inputs.

        :param auto_pause: Whether or not "tabbing out" of the window pauses the game.
        :param use_opengl_if_possible: Whether or not to use OpenGL for rendering, if available.

        """
        if not self._sdl_window:
            self._sdl_window = DeSmuME_SDL_Window(self, auto_pause, use_opengl_if_possible)
        return self._sdl_window

    def display_buffer(self):
        """Return the display buffer in the internal format. You probably want to use display_buffer_as_rgbx instead."""
        self.lib.desmume_draw_raw.restype = POINTER(c_uint16)
        return self.lib.desmume_draw_raw()

    def display_buffer_as_rgbx(self, reuse_buffer=True) -> memoryview:
        """
        Return the display buffer as RGBX color values,
        see the screen size constants for how many pixels make up lines.
        """
        if reuse_buffer:
            if not self._raw_buffer_rgbx:
                self._raw_buffer_rgbx = create_string_buffer(SCREEN_WIDTH * SCREEN_HEIGHT_BOTH * 4)
            buff = self._raw_buffer_rgbx
        else:
            buff = create_string_buffer(SCREEN_WIDTH * SCREEN_HEIGHT_BOTH * 4)
        self.lib.desmume_draw_raw_as_rgbx(cast(buff, c_char_p))
        # XXX: Yes this is stupid, but cairo NEEDS a bytearray....
        return memoryview(bytearray(buff.raw))

    def screenshot(self) -> Image.Image:
        """Convert the current display buffer into a PIL image."""
        buff = create_string_buffer(SCREEN_WIDTH * SCREEN_HEIGHT_BOTH * 3)
        self.lib.desmume_screenshot(cast(buff, c_char_p))

        return Image.frombuffer('RGB', (SCREEN_WIDTH, SCREEN_HEIGHT_BOTH), buff.raw, 'raw', 'RGB', 0, 1)

    def get_ticks(self) -> int:
        """Get the current SDL tick number."""
        return self.lib.desmume_sdl_get_ticks()

    def volume_get(self) -> int:
        """Get the current value (between 0 - 100)."""
        return self.lib.desmume_volume_get()

    def volume_set(self, volume: int):
        """Set the current value (to a value between 0 - 100)."""
        return self.lib.desmume_volume_set(volume)

    def gpu_get_layer_main_enable_state(self, layer_index: int):
        """Get the current display status of the specified layer on the main GPU."""
        return bool(self.lib.desmume_gpu_get_layer_main_enable_state(layer_index))

    def gpu_get_layer_sub_enable_state(self, layer_index: int):
        """Get the current display status of the specified layer on the sub GPU."""
        return bool(self.lib.desmume_gpu_get_layer_sub_enable_state(layer_index))

    def gpu_set_layer_main_enable_state(self, layer_index: int, state: bool):
        """Set the current display status of the specified layer on the main GPU."""
        self.lib.desmume_gpu_set_layer_main_enable_state(layer_index, state)

    def gpu_set_layer_sub_enable_state(self, layer_index: int, state: bool):
        """Set the current display status of the specified layer on the sub GPU."""
        self.lib.desmume_gpu_set_layer_sub_enable_state(layer_index, state)



#  Copyright 2020 Marco Köpcke (Parakoopa)
#
#  This file is part of py-desmume.
#
#  py-desmume is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  py-desmume is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with py-desmume.  If not, see <https://www.gnu.org/licenses/>.
import gettext
import os
from typing import Dict, Optional, List

from gi.repository import Gtk, Gdk

from desmume.controls import key_names, Keys, key_names_localized
from desmume.emulator import DeSmuME_Input
from desmume.frontend.widget_to_primitive import widget_to_primitive
from desmume.i18n_util import _


class JoystickControlsDialogController:
    """This dialog shows the joystick controls."""
    def __init__(self, parent_window: Gtk.Window):
        path = os.path.abspath(os.path.dirname(__file__))
        # SkyTemple translation support
        try:
            from skytemple.core.ui_utils import make_builder
            self.builder = make_builder(os.path.join(path, "PyDeSmuMe_controls.glade"))
        except ImportError:
            self.builder = Gtk.Builder()
            self.builder.add_from_file(os.path.join(path, "PyDeSmuMe_controls.glade"))
        self.window: Gtk.Window = self.builder.get_object('wJoyConfDlg')
        self.window.set_transient_for(parent_window)
        self.window.set_attached_to(parent_window)
        self._joystick_cfg: Optional[List[int]] = None
        self._emulator_input: DeSmuME_Input = None
        self.builder.connect_signals(self)

    def run(self,
            joystick_cfg: List[int],
            emulator_input: DeSmuME_Input,
            emulator_is_running: bool
            ) -> List[int]:
        """Configure the joystick configuration provided using the dialog,
        is is immediately changed in the debugger The new/old (if canceled) config is also returned."""
        self._joystick_cfg = joystick_cfg
        self._emulator_input = emulator_input
        if self._emulator_input.joy_number_connected() < 1 or emulator_is_running:
            if self._emulator_input.joy_number_connected() < 1:
                text = _("You don't have any joypads!")
            else:
                text = _("Can't configure joystick while the game is running!")

            md = Gtk.MessageDialog(None,
                                   Gtk.DialogFlags.DESTROY_WITH_PARENT | Gtk.DialogFlags.MODAL, Gtk.MessageType.ERROR,
                                   Gtk.ButtonsType.OK, text,
                                   title="Error!")
            md.set_position(Gtk.WindowPosition.CENTER)
            md.run()
            md.destroy()
        else:
            for i in range(0, Keys.NB_KEYS):
                b = self.builder.get_object(f"button_joy_{key_names[i]}")
                b.set_label(f"{key_names_localized[i]} : {self._joystick_cfg[i]}")
            self.window.run()
            self.window.hide()

        return self._joystick_cfg

    # KEYBOARD CONFIG / KEY DEFINITION
    def on_wKeyDlg_key_press_event(self, widget: Gtk.Widget, event: Gdk.EventKey, *args):
        pass  # not part of this

    def on_button_kb_key_clicked(self, w, *args):
        pass  # not part of this

    # Joystick configuration / Key definition
    def on_button_joy_key_clicked(self, w, *args):
        key = widget_to_primitive(w)
        dlg = self.builder.get_object("wJoyDlg")
        key -= 1  # key = bit position, start with
        dlg.show_now()
        # Need to force event processing. Otherwise, popup won't show up.
        while Gtk.events_pending():
            Gtk.main_iteration()

        joykey = self._emulator_input.joy_get_set_key(key)

        self._joystick_cfg[key] = joykey

        self.builder.get_object(f"button_joy_{key_names[key]}").set_label(f"{key_names_localized[key]} : {joykey}")

        dlg.hide()

    def gtk_widget_hide_on_delete(self, w: Gtk.Widget, *args):
        w.hide_on_delete()
        return True
